import click
from click.testing import CliRunner
import sys
import pytest
from io import StringIO
from unittest.mock import patch, MagicMock

from snaketool_utils.cli_utils import (
    OrderedCommands,
    echo_click,
    msg,
    msg_box,
    read_config,
    recursive_merge_config,
    write_config,
    update_config,
    copy_config,
    run_snakemake,
    tuple_to_list,
)


def test_list_commands_preserves_order():
    class TestCLI(OrderedCommands):
        pass

    @click.command()
    def command1():
        pass

    @click.command()
    def command2():
        pass

    @click.command()
    def command3():
        pass

    test_cli = TestCLI()
    test_cli.add_command(command1)
    test_cli.add_command(command2)
    test_cli.add_command(command3)

    runner = CliRunner()
    result = runner.invoke(test_cli, ["--help"])
    print(result.output)
    assert result.exit_code == 0
    assert "Commands:\n  command1\n  command2\n  command3" in result.output


def test_echo_click(capsys, tmp_path):
    # Redirect stderr to capture the output
    sys.stderr = StringIO()

    # Define the error message and create a temporary log file
    msg = "This is an error message"
    log_file = tmp_path / "log.txt"

    # Call the echo_click function
    echo_click(msg, log=str(log_file))

    # Get the captured stderr output
    captured_stderr = sys.stderr.getvalue()

    # Assert that the error message is printed to stderr
    assert captured_stderr == msg

    # Assert that the error message is written to the log file
    with open(log_file, "r") as log:
        log_content = log.read()
        assert log_content == msg

    # Clean up and restore stderr
    sys.stderr = sys.__stderr__


def test_msg(capsys, tmp_path):
    # Define the error message and log file path
    err_message = "This is an error message"
    log_file = tmp_path / "log.txt"

    # Patch the strftime function to return a fixed timestamp
    fixed_timestamp = "[2022:01:01 00:00:00] "
    with patch("snaketool_utils.cli_utils.strftime", return_value=fixed_timestamp):
        # Call the msg function
        msg(err_message, log=log_file)

    # Capture the printed output
    captured = capsys.readouterr()

    # Assert that the output contains the formatted error message
    expected_output = fixed_timestamp + err_message + "\n"
    assert captured.err == expected_output

    # Assert that the error message is written to the log file
    with open(log_file, "r") as log:
        log_content = log.read()
        assert log_content == expected_output


def test_msg_box(capsys, tmp_path):
    splash = "Error!"
    errmsg = "This is a multi-line error message\nwith multiple lines"
    log_file = tmp_path / "log.txt"

    msg_box(splash, errmsg=errmsg, log=log_file)

    captured = capsys.readouterr()

    assert splash in captured.err
    assert errmsg in captured.err


@pytest.fixture(scope="function")
def temp_left_config(tmp_path):
    file_path = tmp_path / "config.yaml"
    config_content = """
            key1: value1
            key2: value2
            key3:
                - item1
                - item2
        """

    with open(file_path, "w") as f:
        f.write(config_content)

    expected_config = {"key1": "value1", "key2": "value2", "key3": ["item1", "item2"]}

    yield str(file_path), expected_config


def test_read_config(temp_left_config):
    file_path, expected_config = temp_left_config

    config = read_config(file_path)

    assert config == expected_config


@pytest.fixture(scope="function")
def left_config():
    return {
        "key1": "value1",
        "key2": {"nested_key1": "nested_value1", "nested_key2": "nested_value2"},
        "key3": ["item1", "item2"],
    }


@pytest.fixture(scope="function")
def left_yaml():
    return (
        "key1: value1\n"
        "key2:\n"
        "  nested_key1: nested_value1\n"
        "  nested_key2: nested_value2\n"
        "key3:\n"
        "- item1\n"
        "- item2\n"
    )


@pytest.fixture(scope="function")
def left_path(tmp_path, left_yaml):
    file_path = tmp_path / "left_config.yaml"
    with open(file_path, "w") as f:
        f.write(left_yaml)
    return file_path


@pytest.fixture(scope="function")
def right_config():
    return {
        "key1": "new_value1",
        "key2": {"nested_key2": "new_nested_value2", "nested_key3": "nested_value3"},
        "key4": "value4",
    }


@pytest.fixture(scope="function")
def right_yaml():
    return (
        "key1: value1\n"
        "key2:\n"
        "  nested_key2: new_nested_value2\n"
        "  nested_key3: nested_value3\n"
        "key4: value4\n"
    )


@pytest.fixture(scope="function")
def right_path(tmp_path, right_yaml):
    file_path = tmp_path / "right_config.yaml"
    with open(file_path, "w") as f:
        f.write(right_yaml)
    return file_path


@pytest.fixture(scope="function")
def merged_config():
    return {
        "key1": "new_value1",
        "key2": {
            "nested_key1": "nested_value1",
            "nested_key2": "new_nested_value2",
            "nested_key3": "nested_value3",
        },
        "key3": ["item1", "item2"],
        "key4": "value4",
    }


@pytest.fixture(scope="function")
def merged_yaml():
    return (
        "key1: new_value1\n"
        "key2:\n"
        "  nested_key1: nested_value1\n"
        "  nested_key2: new_nested_value2\n"
        "  nested_key3: nested_value3\n"
        "key3:\n"
        "- item1\n"
        "- item2\n"
        "key4: value4\n"
    )


@pytest.fixture(scope="function")
def merged_path(tmp_path, merged_yaml):
    file_path = tmp_path / "merge_config.yaml"
    with open(file_path, "w") as f:
        f.write(merged_yaml)
    return file_path


def test_recursive_merge_config(left_config, right_config, merged_config):
    recursive_merge_config(left_config, right_config)
    assert left_config == merged_config


def test_write_config(left_config, left_yaml, tmp_path):
    file_path = tmp_path / "config.yaml"
    write_config(left_config, file_path)
    with open(file_path, "r") as f:
        assert f.read() == left_yaml


def test_update_config(tmp_path, left_path, right_config, merged_yaml):
    file_path = tmp_path / "config.yaml"
    update_config(in_config=left_path, merge=right_config, output_config=file_path)
    with open(file_path, "r") as f:
        assert f.read() == merged_yaml


def test_copy_config(tmp_path, left_path, right_config, merged_yaml):
    file_path = tmp_path / "config.yaml"
    copy_config(file_path, merge_config=right_config, system_config=left_path)
    with open(file_path, "r") as f:
        assert f.read() == merged_yaml


def test_run_snakemake(tmp_path):
    # Create temporary files for the configfile, system config, and Snakefile
    configfile = tmp_path / "config.yaml"
    system_config = tmp_path / "system_config.yaml"
    snakefile_path = tmp_path / "Snakefile"
    log_file = tmp_path / "log"

    # Write the system config content to the system config file
    with open(system_config, "w") as f:
        f.write("key: value")

    # Write the Snakefile content to the Snakefile
    with open(snakefile_path, "w") as f:
        f.write("rule all:\n  input: 'output.txt'")

    # Patch the copy_config, update_config, read_config, and subprocess.run functions
    with patch("snaketool_utils.cli_utils.copy_config") as mock_copy_config, patch(
        "snaketool_utils.cli_utils.update_config"
    ) as mock_update_config, patch(
        "snaketool_utils.cli_utils.read_config"
    ) as mock_read_config, patch(
        "subprocess.run"
    ) as mock_run:
        # Set the return values and side effects of the mocked functions
        mock_read_config.return_value = {"key": "value"}

        # Create a MagicMock object to use as the return value of subprocess.run
        mock_run_result = MagicMock()
        mock_run_result.returncode = 0

        # Set the return value of subprocess.run to the MagicMock object
        mock_run.return_value = mock_run_result

        # Call the run_snakemake function
        exit_code = run_snakemake(
            configfile=configfile,
            system_config=system_config,
            snakefile_path=snakefile_path,
        )

        # Assert that the copy_config function was called with the expected arguments
        mock_copy_config.assert_called_once_with(
            configfile, system_config=system_config, log=None
        )

        # Assert that the update_config function was not called
        mock_update_config.assert_not_called()

        # Assert that the read_config function was called with the configfile
        mock_read_config.assert_called_once_with(configfile)

        # Assert that the subprocess.run function was called with the expected command
        mock_run.assert_called_once_with(
            "snakemake -s {} --configfile {} --cores 1".format(
                snakefile_path, configfile
            ),
            shell=True,
        )

        # Assert that the exit code is 0
        assert exit_code == 0

    # Patch the copy_config, update_config, read_config, and subprocess.run functions
    with patch("snaketool_utils.cli_utils.copy_config") as mock_copy_config, patch(
        "snaketool_utils.cli_utils.update_config"
    ) as mock_update_config, patch(
        "snaketool_utils.cli_utils.read_config"
    ) as mock_read_config, patch(
        "subprocess.run"
    ) as mock_run:
        # Set the return values and side effects of the mocked functions
        mock_read_config.return_value = {"key": "value"}

        # Create a MagicMock object to use as the return value of subprocess.run
        mock_run_result = MagicMock()
        mock_run_result.returncode = 0

        # Set the return value of subprocess.run to the MagicMock object
        mock_run.return_value = mock_run_result

        # Call the run_snakemake function again with additional arguments
        exit_code = run_snakemake(
            configfile=configfile,
            system_config=system_config,
            snakefile_path=snakefile_path,
            merge_config={"key2": "value2"},
            threads=4,
            use_conda=True,
            conda_prefix="/path/to/conda",
            snake_default=["--verbose"],
            snake_args=["--dry-run"],
            profile="my_profile",
            log=log_file,
            additional_arg="value",
        )

        # Assert that the copy_config function was called with the expected arguments
        mock_copy_config.assert_called_once_with(
            configfile, system_config=system_config, log=log_file
        )

        # Assert that the update_config function was called with the expected arguments
        # mock_update_config.assert_called_once_with(configfile, merge={"key2": "value2"}, log=log_file)

        # Assert that the read_config function was called with the configfile
        mock_read_config.assert_called_once_with(configfile)

        # Assert that the subprocess.run function was called with the expected command
        expected_command = "snakemake -s {} --configfile {} --use-conda --conda-prefix /path/to/conda --verbose --dry-run --profile my_profile".format(
            snakefile_path, configfile
        )
        mock_run.assert_called_once_with(expected_command, shell=True)

        # Assert that the exit code is 0
        assert exit_code == 0


def test_tuple_to_list_single_tuple():
    input_dict = {'a': (1, 2, 3)}
    expected_output = {'a': [1, 2, 3]}
    assert tuple_to_list(input_dict) == expected_output


def test_tuple_to_list_nested_tuples():
    input_dict = {'a': (1, 2, 3), 'b': {'c': (4, 5)}}
    expected_output = {'a': [1, 2, 3], 'b': {'c': [4, 5]}}
    assert tuple_to_list(input_dict) == expected_output


def test_tuple_to_list_no_tuples():
    input_dict = {'a': 1, 'b': {'c': 2}}
    expected_output = {'a': 1, 'b': {'c': 2}}
    assert tuple_to_list(input_dict) == expected_output


def test_tuple_to_list_empty_dict():
    input_dict = {}
    expected_output = {}
    assert tuple_to_list(input_dict) == expected_output


def test_tuple_to_list_mixed_types():
    input_dict = {'a': (1, 2, 3), 'b': 'string', 'c': {'d': (4, 5)}, 'e': 6}
    expected_output = {'a': [1, 2, 3], 'b': 'string', 'c': {'d': [4, 5]}, 'e': 6}
    assert tuple_to_list(input_dict) == expected_output


from setuptools import setup


def get_version():
    with open("VERSION", "r") as f:
        return f.readline().strip()


def get_description():
    with open("README.md", "r") as f:
        long_description = f.read()
    return long_description


CLASSIFIERS = [
    "Environment :: Console",
    "Environment :: MacOS X",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT license",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS :: MacOS X",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]


setup(
    name="snaketool_utils",
    url="https://github.com/beardymcjohnface/snaketool-utils",
    python_requires=">=3.7",
    description="Utility functions for Snaketool CLI",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    version=get_version(),
    author="Michael Roach",
    author_email="beardymcjohnface@gmail.com",
    install_requires=[
        "pyyaml>=6.0",
        "click>=8.1.3"
    ]
)



import sys
import os
import subprocess
import yaml
import click
import collections.abc
from shutil import copyfile
from time import localtime, strftime


class OrderedCommands(click.Group):
    """This class will preserve the order of subcommands, which is useful when printing --help"""

    def list_commands(self, ctx: click.Context):
        return list(self.commands)


def echo_click(msg, log=None):
    """Print Error message to STDERR and copy to log file

    Args:
        msg (str): Error message to print
        log (srt): Filepath to log file for writing
    """
    click.echo(msg, nl=False, err=True)
    if log:
        with open(log, "a") as l:
            l.write(msg)


def msg(err_message, log=None):
    """Format error message for printing

    Args:
        err_message (str): Error message to print
        log (str): Filepath to log file for writing
    """
    tstamp = strftime("[%Y:%m:%d %H:%M:%S] ", localtime())
    echo_click(tstamp + err_message + "\n", log=log)


def msg_box(splash, errmsg=None, log=None):
    """Fancy formatting and multi-line error message for printing

    Args:
        splash (str): Short splash message to appear in box
        errmsg (str): Long error message to print
        log (str): Filepath to log file for writing
    """
    msg("-" * (len(splash) + 4), log=log)
    msg(f"| {splash} |", log=log)
    msg(("-" * (len(splash) + 4)), log=log)
    if errmsg:
        echo_click("\n" + errmsg + "\n", log=log)


def read_config(file):
    """Read a config file to a dictionary

    Args:
        file (str): Filepath to config YAML file for reading

    Returns (dict): Config read from YAML file
    """

    with open(file, "r") as stream:
        config = yaml.safe_load(stream)
    return config


def recursive_merge_config(config, overwrite_config):
    """Recursively merge a dictionary.

    This is required for updating/merging config dictionaries that are more than one level deep.

    Args:
        config (dict): Config dictionary to overwrite (e.g. defaults)
        overwrite_config (dict): Config dictionary of new/updated values to add
    """

    def _update(d, u):
        for key, value in u.items():
            if isinstance(value, collections.abc.Mapping):
                d[key] = _update(d.get(key, {}), value)
            else:
                d[key] = value
        return d

    _update(config, overwrite_config)


def update_config(in_config=None, merge=None, output_config=None, log=None):
    """Update the default config with the new config values

    Args:
        in_config (str): Filepath to YAML config file
        merge (dict): New values to merge into new config file
        output_config (str): Filepath to write new merged config YAML file
        log (str): Log file for writing STDERR
    """
    if output_config is None:
        output_config = in_config
    config = read_config(in_config)
    msg("Updating config file with new values", log=log)
    recursive_merge_config(config, merge)
    write_config(config, output_config, log=log)


def tuple_to_list(dictionary):
    """Convert click tuples to lists in (nested) dictionaries for safe dumping with pyyaml

    Args:
        dictionary (dict): dictionary of config for Snakemake

    Returns:
        dictionary (dict): dictionary with tuples converted to lists
    """
    out_dict = {}
    for key, value in dictionary.items():
        if isinstance(value, tuple):
            out_dict[key] = list(value)
        elif isinstance(value, dict):
            out_dict[key] = tuple_to_list(value)
        else:
            out_dict[key] = value
    return out_dict


def write_config(config, file, log=None):
    """Write the config dictionary to a YAML file

    Args:
        config (dict): Dictionary of config values
        file (str): Filepath of config file for writing
        log (str): Filepath of log file for writing STDERR
    """
    msg(f"Writing config file to {file}", log=log)
    config = tuple_to_list(config)
    with open(file, "w") as stream:
        yaml.dump(config, stream)


def copy_config(
    local_config,
    merge_config=None,
    system_config=None,
    log=None,
):
    """Copy a config file, optionally merging in new config values.

    Args:
        local_config (str): Filepath of new config YAML for writing
        merge_config (dict): Config values for merging
        system_config (str): Filepath of original config YAML file for reading
        log (str): Filepath of log file for writing STDERR
    """
    if not os.path.isfile(local_config):
        if len(os.path.dirname(local_config)) > 0:
            os.makedirs(os.path.dirname(local_config), exist_ok=True)
        msg(f"Copying system default config to {local_config}", log=log)

        if merge_config:
            update_config(
                in_config=system_config,
                merge=merge_config,
                output_config=local_config,
                log=log,
            )
        else:
            copyfile(system_config, local_config)
    else:
        msg(
            f"Config file {local_config} already exists. Using existing config file.",
            log=log,
        )


def run_snakemake(
    configfile=None,
    system_config=None,
    snakefile_path=None,
    merge_config=None,
    threads=1,
    use_conda=False,
    conda_prefix=None,
    snake_default=None,
    snake_args=[],
    profile=None,
    log=None,
    **kwargs,
):
    """Run a Snakefile!

    Args:
        configfile (str): Filepath of config file to pass with --configfile
        system_config (str): Filepath of system config to copy if configfile not present
        snakefile_path (str): Filepath of Snakefile
        merge_config (dict): Config values to merge with your config file
        threads (int): Number of local threads to request
        use_conda (bool): Snakemake's --use-conda
        conda_prefix (str): Filepath for Snakemake's --conda-prefix
        snake_default (list): Snakemake args to pass to Snakemake
        snake_args (list): Additional args to pass to Snakemake
        profile (str): Name of Snakemake profile
        log (str): Log file for writing STDERR
        **kwargs:

    Returns (int): Exit code
    """

    snake_command = ["snakemake", "-s", snakefile_path]

    # if using a configfile
    if configfile:
        # copy sys default config if not present
        copy_config(configfile, system_config=system_config, log=log)

        if merge_config:
            update_config(in_config=configfile, merge=merge_config, log=log)

        snake_command += ["--configfile", configfile]

        # display the runtime configuration
        snake_config = read_config(configfile)
        msg_box(
            "Runtime config",
            errmsg=yaml.dump(snake_config, Dumper=yaml.Dumper),
            log=log,
        )

    # add threads
    if "--profile" not in snake_args and profile is None:
        snake_command += ["--cores", threads]

    # add conda args if using conda
    if use_conda:
        snake_command += ["--use-conda"]
        if conda_prefix:
            snake_command += ["--conda-prefix", conda_prefix]

    # add snakemake default args
    if snake_default:
        snake_command += snake_default

    # add any additional snakemake commands
    if snake_args:
        snake_command += list(snake_args)

    # allow double-handling of profile
    if profile:
        snake_command += ["--profile", profile]

    # Run Snakemake!!!
    snake_command = " ".join(str(s) for s in snake_command)
    msg_box("Snakemake command", errmsg=snake_command, log=log)
    if not subprocess.run(snake_command, shell=True).returncode == 0:
        msg("ERROR: Snakemake failed", log=log)
        sys.exit(1)
    else:
        msg("Snakemake finished successfully", log=log)
    return 0
